/*___________________________________________________________________________
RFFT2.ASM       ADSP-21364 Radix-2 DIT Real FFT

Performs a radix-2 FFT of length 64 or greater on input data x(n).  This is
accomplished by doing a N/2 (HN) complex FFT of y(n), where Re[y(n)] = x(n) for 
n=0,1,...,HN-1 and Im[y(n)] = x(m) for m=HN,HN+1,...,N-1.  This is followed by 
some calculation to derive X(k) from Y(k).  This method has been refered to as 
"Packing".  Input data is NOT destroyed during the course of this routine.

        N/2     1st half of real normal-ordered input in DM
        N/2     2nd half of real normal-ordered input in DM
        N/2+1   real part of fft stored in DM 
        N/2+1   imag part of fft stored in PM
        N/2     Interleaved Sin and Cos table stored in DM
        N/4     Cos table stored in PM 
        N/4     Sin table stored in DM

Note that only (N/2+1) elements are necessary to define the real FFT spectra
because of its Hermitian (complex conjugate) symmetry.

Based on FFTRAD2 by Kapriel Karagozian  Analog Devices DSP Div. 1-800-ANALOGD
Author:   25-APR-91 Ronnin Yee          Analog Devices DSP Div. 1-800-ANALOGD

Modified:   11/98   Richard Grafton     Analog Devices DSP Div. 1-800-ANALOGD
                                        (converted for ADSP-21364) 
Modified:    4/99   Richard Grafton     Analog Devices DSP Div. 1-800-ANALOGD
                                        (optimized for SIMD of ADSP-21364)
Modified:   6/00    Philip Giordano     Resolved illegal mixed word dual-data SIMD access
                                        extra cycle introduced to resolve access
                                        (must further optimize this access as well as
                                        final stage which is not SIMD optimized at all)
                                        Code verified on Rev 0.1 silicon and 4.1.2 tools release,
Modified:   9/04    Joe Tarkoff 	    Ported to VDSP++ 3.5
                                        Replaced .segment with .section format


Calling Information:
    dm(twiddle[N/2])    - Interleaved sin(2pi*n/HN) table and cos(2pi*n/HN) table from RFFT2TBL program
                          twiddle factors interleaved as sin[0], cos[0], sin[1], cos[1] ...
                          stored in long-word space
    pm(h_cosine[N/4])   - 0.5*cos(pi*n/HN) table from RFFT2TBL program
    dm(h_sine[N/4)      - 0.5*sin(pi*n/HN) table from RFFT2TBL program
    dm(real[N/2])       - 1st half of real input array stored in dm and treated as real
    pm(imag[N/2])       - 2nd half of real input array stored in pm and treated as imag

    (Note:  Because the bit reversed address mode is used with the array
    real and the array imag, they must start at addresses that are integer multiples of
    the length (N) of the transform, (i.e.  0,N,2N,3N,...).  This is accomplished by
    specifying the segment starting at that addresses in the architecture file
    and placing the variable alone in its own segment.  These addresses must
    also be reflected in the preprocessor variables IREDM and IREPM respectively
    in bit reversed format.)

    The data files, "thc.dat" and "ths.dat" are cosine and sine coefficients 
    used in the final convolution stage.

    They are derived using the following equations where n=1:N/4 
        h_cosine(n)   0.5*cos(pi*n/HN)
        h_sine(n)     0.5*sin(pi*n/HN)
    
    where HN=N/2 and N is the point size of the fft used.


Results:
    dm(refft[N/2+1])   - real working array and output
    pm(imfft[N/2+1])   - imaginary working array and output

Benchmarks: Radix-2, real, from call to return

                                  time (us)                      time (us)
    FFT Length    cycles (SISD)    100MHz       cycles (SIMD)     100MHz  
    ----------    ------          --------      ------           -------- 
                                                                         
         64         636              6.36         486               4.86  
        128        1231             12.31         868               8.68  
        256        2504             25.04        1664              16.64  
        512        5286             52.86        3357              33.57  
       1024       11331            113.31        6970              69.70  
       2048       24418            244.18       14679             146.79  
       4096       52608            526.08       31092             310.92  
       8192      113054           1130.54       65936             659.36  
                                                                
      

   First 2 Stages    -  5 cycles per 4 butterflies
   Middle Stages     -  4 cycles per 2 butterflies
   2nd to Last Stage - 10 cycles per 4 butterflies
   Last FFT Stage    -  6 cycles per 2 butterflies
   Conversion Stage  -  6 cycles per 2 elements (N/2 elements total)

Memory Usage:
    pm code = 187 words, pm data = N+1 words, dm data = 2*N+1 words
____________________________________________________________________________*/

/* Include for symbolic definition of system register bits */
#include "def21364.h"

/*_________The constants below must be changed for different length FFTs______
N         = number of points in the FFT, must be a power of 2
STAGES    = log2(N)
BRMODIFY  = bitrev(32 bit N/4)
____________________________________________________________________________*/
#define N               128             /* value must be modified if length of FFT computation is changed */
#define HN              N/2
#define STAGES          7               /* value must be modified if length of FFT computation is changed */        
#define BRMODIFY        0x04000000      /* value must be modified if length of FFT computation is changed */

/*________These constants are independent of the number of points____________*/
/* files for twiddle factors and sin/cos coeffs must be modified if length of FFT computation is changed */
#define BFLY8           4               /*Offset between bf branches in a group of 8*/

.SECTION/DM             seg_dm64;
.VAR    twiddle[HN]     =   "twiddle.dat";  /* twiddle factors, from RFFT2TBL, interleaved as   */
                                            /* sin[0], cos[0], sin[1], cos[1] ...               */


.SECTION/DM             seg_dmda;
.ALIGN 2;
.VAR    h_sine[HN/2]    =   "ths.dat";      /* end conv. sin coeff. from RFFT2TBL*/
.VAR    refft[HN+2];                        /* real result */
.GLOBAL refft;


.SECTION/DM             seg_dmrd;       /* Segment Addr. = integer multiple of N */
.ALIGN 2;
.VAR    real[HN]        =   "inreal1.dat";  /* 1st half of input real array treated as real */
.GLOBAL real;


.SECTION/DM             seg_pmda;
.ALIGN 2;
.VAR    h_cosine[HN/2]  =   "thc.dat";      /* end conv. cos coeff. from RFFT2TBL*/
.VAR    imfft[HN+2];                        /* imag result */
.GLOBAL imfft;


.SECTION/DM             seg_pmrd;
.ALIGN 2;
.VAR    imag[HN]        =   "inreal2.dat";  /* 2nd half of input real array treated as imag*/
.GLOBAL imag;


/*______________________reset vector test call of fft______________________*/
.SECTION/PM             seg_rth;        /* program starts at the reset vector */

Resrvd: nop;nop;nop;nop;
        
Reset:  idle;
        call rfft2;
stop:   idle;
        nop;



.SECTION/PM             seg_pmco;
/*______________________________begin FFT__________________________________*/

rfft2:  f4=-1.0;
        s4=1.0;
        
        bit set MODE1 BR0 | BR8 | RND32 | CBUFEN | PEYEN;       /* enable bit reverse of i0 */

/*Do bitrev and packing within first two stages*/
        i0=real;
        bitrev(i0,0);
		b0=i0;
		l0=0;
        m0=BRMODIFY;
        
        i8=imag;
        bitrev(i8,0);
        b8=i8;
        l8=0;
        m8=BRMODIFY;
        
        b1=real+HN;
        l1=0;
        m1=2;

        r0=refft;
        r0=lshift r0 by -1;
        b2=r0;
        l2=HN;
        m1=1;                       

        r0=imfft;
        r0=lshift r0 by -1;
        b10=r0;
        l10=HN;
        m9=1;

        /*Do the first two stages (actually a radix-4 FFT stage)*/

                                                                f8=dm(i0,m0),   f9=pm(i8,m8);
                                f8=f8+f9,       f9=f8-f9,       f10=dm(i0,m0),  f11=pm(i8,m8);
                                f12=f10+f11,    f3=f10-f11;
                f3=f3*f4,       f2=f8+f12,      f6=f8-f12;
                
LCNTR=HN/4,     do FSTAGE until LCE;        /* do N/4 simple radix-4 butterflies */
                                                                r3<->s3;
                                f3=f9+f3,       f7=f9-f3,       f8=dm(i0,m0),   f9=pm(i8,m8);
                                f8=f8+f9,       f9=f8-f9,       f10=dm(i0,m0),  f11=pm(i8,m8);
                                f12=f10+f11,    f3=f10-f11,     dm(i2,m1)=f2,   pm(i10,m9)=f2;
FSTAGE:         f3=f3*f4,       f2=f8+f12,      f6=f8-f12,      dm(i2,m1)=f6,   pm(i10,m9)=f6;
         


/*middle stages loop */

        bit clr MODE1 BR0 | BR8;        /*finished with bitreversal*/
        b2=refft;
        b8=imfft;
        l8=HN;
        
        b0=refft;
        l0=HN;
    
        b1=twiddle;
        l1=@twiddle/2;

        b10=imfft;
        l10=HN;
        
        b11=imfft;
        l11=HN;

        m0=-BFLY8;
        m1=-HN/8;
        m2=-BFLY8-2;
        m11=-2;

        r2=2;
        r3=-BFLY8;              /*initializes m0,10 - incr for butterf branches */
        r5=BFLY8/2;             /*counts # butterflies per a group              */
        r9=(-2*BFLY8)-2;        /*initializes m12 - wrap around to next grp + 2 */
        r10=-2*BFLY8;           /*initializes m8 - incr between groups          */
        r13=-BFLY8-2;           /*initializes m2,13 - wrap to bgn of 1st group  */
        r15=HN/8;               /*# OF GROUPS IN THIRD STAGE                    */

        f1=dm(i1,m1);           /*set pointers to tables to 1st coeff.   */
        
LCNTR=STAGES-5, do end_stage until LCE; /*# OF STAGES TO BE HANDLED = LOG2(HN)-4 */
        m8=r10;
        m10=r3;
        m12=r9;
        i0=refft+HN-2;
        i2=refft+HN-2;
        i8=imfft+HN-2;
        i10=imfft+HN-2;
        i11=imfft+HN-2;
        r15=r15-r2,     m13=r13;        /*CALCULATE # OF CORE  */
                                        /*BFLIES/GROUP IN THIS STAGE */

                                        s0=dm(i1,m1);

                                        r0=dm(i1,m1); f7=pm(i8,m8); /* resolved illegal mixed word dual-data SIMD access */
f12=f0*f7,                              f6=dm(i0,m0);
f8=f1*f6,                                               modify(i11,m10);
f11=f1*f7;                                              
f14=f0*f6,  f12=f8+f12,                 f8=dm(i0,m0),   f7=pm(i8,m8);
f12=f0*f7,  f13=f8+f12, f10=f8-f12,     f6=dm(i0,m0);   

/*Each iteration does another set of bttrflys in each group */

LCNTR=r5,   do end_group until LCE;     /*# OF BUTTERFLIES/GROUP IN THIS STAGE */

/*core butterfly loop */

LCNTR=r15,  do end_bfly until LCE;      /*Do a butterfly in each group - 2 */
    f8=f1*f6,   f14=f11-f14,                    dm(i2,m0)=f10,  f9=pm(i11,m8);
    f11=f1*f7,  f3=f9+f14,      f9=f9-f14,      dm(i2,m0)=f13,  f7=pm(i8,m8);
    f14=f0*f6,  f12=f8+f12,                     f8=dm(i0,m0),   pm(i10,m10)=f9;
end_bfly:
    f12=f0*f7,  f13=f8+f12,     f10=f8-f12,     f6=dm(i0,m0),   pm(i10,m10)=f3;

/*finish up last bttrfly and set up for next butterfly in each group */

f8=f1*f6,               f14=f11-f14,    dm(i2,m0)=f10,  f9=pm(i11,m8);
f11=f1*f7,  f4=f9+f14,  f9=f9-f14,      dm(i2,m0)=f13,  f14=pm(i8,m11);
f14=f0*f6,  f12=f8+f12,                 f8=dm(i0,m2),   pm(i10,m10)=f9;
                                        s0=dm(i1,m1);
            f13=f8+f12, f10=f8-f12,     r0=dm(i1,m1);   f7=pm(i8,m8);   /* resolved illegal mixed word dual-data SIMD access */
                        f14=f11-f14,    dm(i2,m0)=f10,  f9=pm(i11,m12);

/*start on next butterfly in each group */
f12=f0*f7,  f3=f9+f14,  f9=f9-f14,      f6=dm(i0,m0);
f8=f1*f6,                               dm(i2,m2)=f13,  pm(i10,m10)=f4;
f11=f1*f7,                                              pm(i10,m10)=f9;
f14=f0*f6,  f12=f8+f12,                 f8=dm(i0,m0),   f7=pm(i8,m8);
end_group:
f12=f0*f7,  f13=f8+f12, f10=f8-f12,     f6=dm(i0,m0),   pm(i10,m13)=f3;

                                        
            r4=r15+r2,          i1=b1;              /*PREPARE R4 FOR #OF BFLIES CALC */
            r15=ashift r4 by -1;                    /*# OF BFLIES/GRP IN NEXT STAGE */
            r4=-r15;
            m1=r4;                                  /*update inc for sin & cos  */
            r5=ashift r5 by 1,  f1=dm(i1,m1);       /*update # bttrfly in a grp */
            r3=ashift r5 by 1;
            r3=-r3;                                 /* inc for bttrfly branch */
            r13=r3-r2,           m0=r3;             /* wrap to 1st grp        */
            r10=ashift r3 by 1;                     /* inc between grps       */
end_stage:  r9=r10-r2,           m2=r13;            /* wrap to grp +1         */

/*_________ next to last stage__________*/
        m1=-2;                  /*modifier to sine table pntr    */
        m8=r10;                 /*incr between groups            */
        m10=r3;                 /*incr between bttrfly branches  */
        m12=r9;                 /*wrap around to next grp + 1    */
        m13=r13;                /*wrap to bgn of 1st group       */

        i0=refft+HN-2;
        
        r0=b1;
        r1=(HN/2)-2;
        r0=r0+r1;
        i1=r0;                  /*pntr to 1st twiddle coeffs     */
        
        i2=refft+HN-2;
        i8=imfft+HN-2;
        i10=imfft+HN-2;
        i11=imfft+HN-2;

                                        s0=dm(i1,m1);

                                        r0=dm(i1,m1);   f7=pm(i8,m8);  /* resolved illegal mixed word dual-data SIMD access */
f12=f0*f7,                              f6=dm(i0,m0);
f8=f1*f6,                               modify(i11,m10);
f11=f1*f7,                                              f7=pm(i8,m12);
f14=f0*f6, f12=f8+f12,                  f8=dm(i0,m0);
f12=f0*f7, f13=f8+f12,  f10=f8-f12,     f6=dm(i0,m0);                           

/*Do the HN/4 butterflies in the two groups of this stage */

LCNTR=HN/8,  do end_group2 until LCE;   
    f8=f1*f6,                   f14=f11-f14,    dm(i2,m0)=f10,  f9=pm(i11,m8);
    f11=f1*f7,  f3=f9+f14,      f9=f9-f14,      dm(i2,m0)=f13;
    f14=f0*f6,  f12=f8+f12,                     f8=dm(i0,m2),   pm(i10,m10)=f9;
                                                s0=dm(i1,m1);
                f13=f8+f12,     f10=f8-f12,     r0=dm(i1,m1);   f7=pm(i8,m8);  /* resolved illegal mixed word dual-data SIMD access */

    f12=f0*f7,                  f14=f11-f14,    f6=dm(i0,m0),   f9=pm(i11,m12);

    f8=f1*f6,   f3=f9+f14,      f9=f9-f14,      dm(i2,m0)=f10,  pm(i10,m10)=f3;
    f11=f1*f7,                                  dm(i2,m2)=f13,  pm(i10,m10)=f9; 
    f14=f0*f6, f12=f8+f12,                      f8=dm(i0,m0),   f7=pm(i8,m12);
end_group2:
    f12=f0*f7, f13=f8+f12,      f10=f8-f12,     f6=dm(i0,m0),   pm(i10,m13)=f3;


/*   The last stage      */

        m0=-HN/2;               
        m2=-HN/2-2;
        m10=m0;
        m13=m2;
        i0=refft+HN-2;
        
        r0=b1;
        r1=(HN/2)-1;
        r0=r0+r1;
        i1=r0;                  /*pntr to 1st twiddle coeffs     */

        i2=refft+HN-2;
        i8=imfft+HN-2;
        
        i10=imfft+HN-2;
        i11=imfft+HN-2;
        m1=-1;                  /*modifiers to coeff tables      */

/*start first bttrfly */
                                                        s0=dm(i1,m1);

                                        r0=dm(i1,m1);   f7=pm(i8,m11);  /* resolved illegal mixed word dual-data SIMD access */
f12=f0*f7,                              f6=dm(i0,m0);                                    
f8=f1*f6,                               modify(i11,m10);
f11=f1*f7;                                      
f14=f0*f6, f12=f8+f12,                  f8=dm(i0,m2),   f9=pm(i11,m11);

/*do HN/2 bttrflys in the last stage */

LCNTR=HN/4,  do last_stage until LCE;   
                                                s0=dm(i1,m1);
                f13=f8+f12,     f10=f8-f12,     r0=dm(i1,m1);   f7=pm(i8,m11);  /* resolved illegal mixed word dual-data SIMD access */
    f12=f0*f7,  f14=f11-f14,                    f6=dm(i0,m0);
    f8=f1*f6,   f3=f9+f14,      f15=f9-f14,     dm(i2,m0)=f10,  f9=pm(i11,m11);
    f11=f1*f7,                                  dm(i2,m2)=f13,  pm(i10,m10)=f15;
last_stage:                                             
    f14=f0*f6, f12=f8+f12,                      f8=dm(i0,m2),   pm(i10,m13)=f3;
    
    bit clr mode1 PEYEN;
    m9=-1;

/*compute the N point FFT of real from the packed N/2 transform */

i0=refft;           /* real src pntr forward */
b1=h_sine+1;        /* real sine pntr*/
i2=refft+HN-1;      /* real src pntr backward*/
i3=i0;              /* real dest pntr forward*/
i4=refft+HN;        /* real dest pntr backward*/
l4=0;

i8=imfft;           /* imag src pntr forward*/
b9=h_cosine+1;      /* imag cosine pntr*/
i10=imfft+HN-1;     /* imag src pntr backward*/
i11=i8;             /* imag dest pntr forward*/
i12=imfft+HN;       /* imag dest pntr backward */
l12=0;

m0 = 1;
m8 = 1;
f2 = 0.5;           /* this register is used to div by 2 */
f15= 0.0;

/* do first pair, U(0)+jV(0) & U(HN)+jV(HN) */
                                                f12=dm(i0,m0),  f9 =pm(i8,m8);
                    f11=f12+f9, f3=f12-f9,      f12=dm(i0,m0),  f9 =pm(i8,m8);
                    f10=PASS f15,               f8 =dm(i2,m1),  f13=pm(i10,m9);
                                                dm(i3,m0)=f11;

/* do the rest in pairs (n,HN-n) except the mid-point */

                    f4=f8+f12,  f5=f8-f12,      f0 =dm(i1,m0),  f1 =pm(i9,m8);

LCNTR = (HN/2)-1, DO trans UNTIL LCE;
        f8=f0*f5,   f6=f9+f13,      f7=f9-f13;                      
        f12=f1*f6,  f10=f10+f15,    f15=f10-f15,    dm(i4,m1)=f3,   f13=pm(i10,m9);
        f11=f2*f4,  f14=f8+f12,                     f12=dm(i0,m0),  pm(i11,m8)=f10;
        f10=f1*f5,  f11=f11+f14,    f3=f11-f14,     f8=dm(i2,m1),   pm(i12,m9)=f15;
        f14=f0*f6,  f4=f8+f12,      f5=f8-f12,      f0=dm(i1,m0),   f1=pm(i9,m8);
trans:  f15=f2*f7,  f10=f10-f14,                    dm(i3,m0)=f11,  f9=pm(i8,m8);
                    
/* finish up and do the mid-point */
                    f10=f10+f15,    f15=f10-f15,    dm(i4,m1)=f3;
                    f14=-f9,                                        pm(i11,m8)=f10;
    rts (db);   /* finished */
                                                                    pm(i12,m9)=f15;
                                                    dm(i3,m0)=f8,   pm(i11,m8)=f14;

/*_______________________________________________________________________*/

